#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .base import TemplateWidget
from .buttons import ModalButton, CollapseButton, LinkButton, FormButton

class BasicHeader(TemplateWidget):
    template_name = 'wildewidgets/header_with_controls.html'
    header_level = 1
    header_type = 'h'
    header_text = None
    css_class = "my-3"
    css_id = None
    badge_text = None
    badge_class = "warning"

    def __init__(self, **kwargs):
        self.header_level = kwargs.get('header_level', self.header_level)
        self.header_type = kwargs.get('header_type', self.header_type)
        self.header_text = kwargs.get('header_text', self.header_text)
        self.css_class = kwargs.get('css_class', self.css_class)
        self.css_id = kwargs.get('css_id', self.css_id)
        self.badge_text = kwargs.get('badge_text', self.badge_text)
        self.badge_class = kwargs.get('badge_class', self.badge_class)

    def get_context_data(self, **kwargs):
        if self.header_type == 'h':
            kwargs['header_class'] = f"h{self.header_level}"
        elif self.header_type == 'display':
            kwargs['header_class'] = f"display-{self.header_level}"

        kwargs['header_level'] = self.header_level
        kwargs['header_text'] = self.header_text
        kwargs['header_type'] = self.header_type
        kwargs['css_class'] = self.css_class
        kwargs['css_id'] = self.css_id
        kwargs['badge_text'] = self.badge_text
        kwargs['badge_class'] = self.badge_class
        return kwargs


class HeaderWithLinkButton(BasicHeader):
    template_name = 'wildewidgets/header_with_link_button.html'
    url = None
    link_text = None
    button_class = "primary"

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs['url'] = self.url
        kwargs['link_text'] = self.link_text
        kwargs['button_class'] = self.button_class
        return kwargs


class HeaderWithFormButton(BasicHeader):
    template_name = 'wildewidgets/header_with_form_button.html'
    url = None
    button_text = None

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs['url'] = self.url
        kwargs['button_text'] = self.link_text
        return kwargs


class HeaderWithCollapseButton(BasicHeader):
    template_name = 'wildewidgets/header_with_collapse_button.html'
    collapse_id = None
    button_text = None
    button_class = "primary"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.collapse_id = kwargs.get('collapse_id', self.collapse_id)
        self.button_text = kwargs.get('button_text', self.button_text)
        self.button_class = kwargs.get('button_class', self.button_class)

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs['collapse_id'] = self.collapse_id
        kwargs['button_text'] = self.button_text
        kwargs['button_class'] = self.button_class
        return kwargs


class HeaderWithModalButton(BasicHeader):
    template_name = 'wildewidgets/header_with_modal_button.html'
    modal_id = None
    button_text = None
    button_class = "primary"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.modal_id = kwargs.get('modal_id', self.modal_id)
        self.button_text = kwargs.get('button_text', self.button_text)
        self.button_class = kwargs.get('button_class', self.button_class)

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs['modal_id'] = self.modal_id
        kwargs['button_text'] = self.button_text
        kwargs['button_class'] = self.button_class
        return kwargs


class HeaderWithWidget(BasicHeader):
    template_name = 'wildewidgets/header_with_widget.html'

    def __init__(self, widget=None, **kwargs):
        super().__init__(**kwargs)
        self.widget = widget

    def add_form_button(self, **kwargs):
        widget = FormButton(**kwargs)
        self.widget = widget

    def add_link_button(self, **kwargs):
        widget = LinkButton(**kwargs)
        self.widget = widget

    def add_modal_button(self, **kwargs):
        widget = ModalButton(**kwargs)
        self.widget = widget

    def add_collapse_button(self, **kwargs):
        widget = CollapseButton(**kwargs)
        self.widget = widget

    def set_widget(self, widget):
        self.widget = widget

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs['widget'] = self.widget
        return kwargs


class PageHeader(HeaderWithWidget):
    css_class = "my-4"


class CardHeader(HeaderWithWidget):
    css_class = "my-3"
    header_level = 2



#!/usr/bin/env python
# -*- coding: utf-8 -*-


try:
    from pygments import highlight
    from pygments.lexers import get_lexer_by_name
    from pygments.formatters import HtmlFormatter
except ModuleNotFoundError:
    # Only needed if using syntax highlighting
    pass

from django.core.exceptions import ImproperlyConfigured

from .base import TemplateWidget, Block


class CodeWidget(TemplateWidget):
    template_name = 'wildewidgets/code_widget.html'
    language = None
    code = ""
    Line_numbers = False
    css_class = None

    def __init__(self, *args, **kwargs):
        if 'code' in kwargs:
            self.code = kwargs['code']
        if 'language' in kwargs:
            self.language = kwargs['language']

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        if not self.language:
            raise ImproperlyConfigured("You must define a language.")
        lexer = get_lexer_by_name(self.language)
        formatter = HtmlFormatter(linenos=self.Line_numbers, cssclass="wildewidgets_highlight")
        kwargs['code'] = highlight(self.code, lexer, formatter)
        kwargs['css_class'] = self.css_class
        return kwargs


class MarkdownWidget(TemplateWidget):
    template_name = 'wildewidgets/markdown_widget.html'
    text = ""
    css_class = ""

    def __init__(self, *args, **kwargs):
        self.text = kwargs.get("text", self.text)
        self.css_class = kwargs.get("css_class", self.css_class)

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs['text'] = self.text
        kwargs['css_class'] = self.css_class
        return kwargs


class HTMLWidget(TemplateWidget):
    template_name = 'wildewidgets/html_widget.html'
    html = ""
    css_class = None

    def __init__(self, *args, **kwargs):
        self.html = kwargs.get('html', self.html)

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs['html'] = self.html
        kwargs['css_class'] = self.css_class
        return kwargs


class StringBlock(Block):

    def __init__(self, text: str, **kwargs):
        super().__init__(*[text], **kwargs)


class TimeStamp(StringBlock):
    tag="small"
    css_class="text-muted"


class LabelBlock(StringBlock):
    css_class="font-weight-bold"


class TagBlock(StringBlock):

    def __init__(self, text: str, color: str = "secondary", **kwargs):
        css_class = kwargs.get("css_class", "")
        css_class += f" badge bg-{color}"
        kwargs["css_class"] = css_class.strip()
        super().__init__(text, **kwargs)




import random

from django import template
from django.urls import reverse_lazy

from wildewidgets.views import WidgetInitKwargsMixin


class BasicMenu(WidgetInitKwargsMixin):
    template_file = "wildewidgets/menu.html"
    navbar_classes = "navbar-expand-lg navbar-light"
    container = "container-lg"
    brand_image = None
    brand_image_width = "100%"
    brand_text = None
    brand_url = "#"
    items = []

    def __init__(self, *args, **kwargs):
        self.menu = {}
        self.active = None
        if args:
            self.active_hierarchy = args[0].split('/')
        else:
            self.active_hierarchy

    def build_menu(self):
        if len(self.active_hierarchy) > 0:
            for item in self.items:
                data = {}
                if type(item[1]) == str:
                    data['url'] = reverse_lazy(item[1])
                    data['extra'] = ''
                    data['kind'] = 'item'

                    if len(item) > 2:
                        extra = item[2]
                        if type(extra) == dict:
                            extra_list = []
                            for k, v in extra.items():
                                extra_list.append(f"{k}={v}")
                            extra = f"?{'&'.join(extra_list)}"
                            data['extra'] = extra
                elif type(item[1]) == list:
                    if len(self.active_hierarchy) > 1:
                        submenu_active = self.active_hierarchy[1]
                    else:
                        submenu_active = None
                    data = self.parse_submemu(item[1], submenu_active)

                self.add_menu_item(item[0], data, item[0] == self.active_hierarchy[0])

    def add_menu_item(self, title, data, active=False):
        self.menu[title] = data
        if active:
            self.active = title

    def parse_submemu(self, items, submenu_active):
        data = {
            'kind': 'submenu'
        }
        sub_menu_items = []
        for item in items:
            if not type(item) == tuple:
                continue
            if item[0] == 'divider':
                subdata = {
                    'divider': True
                }
            else:
                subdata = {
                    'title': item[0],
                    'url': reverse_lazy(item[1]),
                    'extra': '',
                    'divider': False,
                    'active': item[0] == submenu_active,
                }

            if len(item) > 2:
                subdata['extra'] = self.convert_extra(item[2])
            sub_menu_items.append(subdata)

        data['items'] = sub_menu_items
        return data

    def get_content(self, **kwargs):
        self.build_menu()
        context = {
            'menu': self.menu,
            'active': self.active,
            'navbar_classes': self.navbar_classes,
            'navbar_container': self.container,
            'brand_image': self.brand_image,
            'brand_image_width': self.brand_image_width,
            'brand_text': self.brand_text,
            'brand_url': self.brand_url,
            'vertical': "navbar-vertical" in self.navbar_classes,
            'target': random.randrange(0, 10000),
        }
        html_template = template.loader.get_template(self.template_file)
        content = html_template.render(context)
        return content

    def __str__(self):
        return self.get_content()


class DarkMenu(BasicMenu):
    navbar_classes = "navbar-expand-lg navbar-dark bg-secondary"


class VerticalDarkMenu(BasicMenu):
    navbar_classes = "navbar-vertical navbar-expand-lg navbar-dark"


class LightMenu(BasicMenu):
    navbar_classes = "navbar-expand-lg navbar-light"


class MenuMixin:
    menu_class = None
    submenu_class = None

    def get_menu_class(self):
        if self.menu_class:
            return self.menu_class
        return None

    def get_menu(self):
        menu_class = self.get_menu_class()
        if menu_class:
            menu_item = None
            if self.menu_item:
                # FIXME: menu_item is not used
                menu_item = self.menu_item
            return menu_class(self.menu_item)
        return None

    def get_submenu_class(self):
        if self.submenu_class:
            return self.submenu_class
        return None

    def get_submenu(self):
        submenu_class = self.get_submenu_class()
        if submenu_class:
            submenu_item = None
            if self.submenu_item:
                submenu_item = self.submenu_item
            return submenu_class(submenu_item)
        return None

    def get_context_data(self, **kwargs):
        menu = self.get_menu()
        submenu = self.get_submenu()
        if menu:
            kwargs['menu'] = menu
        if submenu:
            kwargs['submenu'] = submenu
        return super().get_context_data(**kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .base import TemplateWidget

from django.core.exceptions import ImproperlyConfigured


class ModalWidget(TemplateWidget):
    template_name = 'wildewidgets/modal.html'
    modal_id = None
    modal_title = None

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs['modal_id'] = self.modal_id
        kwargs['modal_title'] = self.modal_title
        return kwargs


class CrispyFormModalWidget(ModalWidget):
    template_name = 'wildewidgets/crispy_form_modal.html'
    form_class = None
    form = None

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        if self.form_class:
            kwargs['form'] = self.form_class()
        elif self.form:
            kwargs['form'] = self.form
        else:
            raise ImproperlyConfigured("Either 'form_class' or 'form' must be set")
        return kwargs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import math
import random

from django import template
from django.conf import settings

from wildewidgets.views import WidgetInitKwargsMixin, JSONDataView

from ..base import Widget


class CategoryChart(Widget, WidgetInitKwargsMixin, JSONDataView):

    COLORS = [
        (0, 59, 76),
        (0, 88, 80),
        (100, 75, 120),
        (123, 48, 62),
        (133, 152, 148),
        (157, 174, 136),
        (159, 146, 94),
        (242, 211, 131),
        (30, 152, 138),
        (115, 169, 80)
    ]

    GRAYS = [
        (200, 200, 200),
        (229, 229, 229),
        (170, 169, 159),
        (118, 119, 123),
        (97, 98, 101),
        (175, 175, 175),
        (105, 107, 115),
    ]
    template_file = 'wildewidgets/categorychart.html'
    legend = False
    legend_position = "top"
    color = True

    def __init__(self, *args, **kwargs):
        self.options = {
            'width':  kwargs.get('width', '400px'),
            'height':  kwargs.get('height', '400px'),
            "title": kwargs.get('title', None),
            "legend": kwargs.get('legend', self.legend),
            "legend_position": kwargs.get('legend_position', self.legend_position),
            "chart_type": kwargs.get('chart_type', None),
            "histogram": kwargs.get('histogram', False),
            "max": kwargs.get('max', None),
            "thousands": kwargs.get('thousands', False),
            "histogram_max": kwargs.get('histogram_max', None),
            "url": kwargs.get('url', None)
        }
        self.chart_id = kwargs.get('chart_id', None)
        self.categories = None
        self.datasets = []
        self.dataset_labels = []
        self.color = kwargs.get('color', self.color)
        self.colors = []
        if hasattr(settings, 'CHARTJS_FONT_FAMILY'):
            self.options['chartjs_font_family'] = settings.CHARTJS_FONT_FAMILY
        if hasattr(settings, 'CHARTJS_TITLE_FONT_SIZE'):
            self.options['chartjs_title_font_size'] = settings.CHARTJS_TITLE_FONT_SIZE
        if hasattr(settings, 'CHARTJS_TITLE_FONT_STYLE'):
            self.options['chartjs_title_font_style'] = settings.CHARTJS_TITLE_FONT_STYLE
        if hasattr(settings, 'CHARTJS_TITLE_PADDING'):
            self.options['chartjs_title_padding'] = settings.CHARTJS_TITLE_PADDING
        super().__init__(*args, **kwargs)

    def set_categories(self, categories):
        self.categories = categories

    def add_dataset(self, dataset, label=None):
        self.datasets.append(dataset)
        self.dataset_labels.append(label)

    def set_option(self, name, value):
        self.options[name] = value

    def set_color(self, color):
        self.color = color

    def set_colors(self, colors):
        self.colors = colors

    def get_content(self, **kwargs):
        if self.chart_id:
            chart_id = self.chart_id
        else:
            chart_id = random.randrange(0, 1000)
        template_file = self.template_file
        if self.datasets:
            context = self.get_context_data()
        else:
            context = {"async": True}
        html_template = template.loader.get_template(template_file)
        context['options'] = self.options
        context['name'] = f"chart_{chart_id}"
        context["wildewidgetclass"] = self.__class__.__name__
        context["extra_data"] = self.get_encoded_extra_data()
        content = html_template.render(context)
        return content

    def __str__(self):
        return self.get_content()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        self.load()
        context.update({"labels": self.get_categories(), "datasets": self.get_dataset_configs()})
        return context

    def get_color_iterator(self):
        if self.colors:
            return iter(self.colors)
        elif self.color:
            return iter(self.COLORS)
        else:
            return iter(self.GRAYS)

    def get_dataset_options(self, index, color):
        default_opt = {
            "backgroundColor": "rgba(%d, %d, %d, 0.5)" % color,
            "borderColor": "rgba(%d, %d, %d, 1)" % color,
            "borderWidth": 0.2,
            # "pointBackgroundColor": "rgba(%d, %d, %d, 1)" % color,
            # "pointBorderColor": "#fff",
        }
        return default_opt

    def get_dataset_configs(self):
        return []

    def get_categories(self):
        if self.categories:
            return self.categories
        raise NotImplementedError(  # pragma: no cover
            "You should return a labels list. " '(i.e: ["January", ...])'
        )

    def get_datasets(self):
        if self.datasets:
            return self.datasets
        raise NotImplementedError(  # pragma: no cover
            "You should return a data list list. " "(i.e: [[25, 34, 0, 1, 50], ...])."
        )

    def get_dataset_labels(self):
        return self.dataset_labels

    def load(self):
        pass


class DoughnutChart(CategoryChart):

    def __init__(self, *args, **kwargs):
        if 'chart_type' not in kwargs:
            kwargs['chart_type'] = "doughnut"
        super().__init__(*args, **kwargs)

    def get_dataset_configs(self):
        datasets = []
        color_generator = self.get_color_iterator()
        data = self.get_dataset()
        dataset = {"data": data}
        dataset["backgroundColor"] = []
        for j in range(len(data)):
            color = tuple(next(color_generator))
            dataset['backgroundColor'].append("rgba(%d, %d, %d)" % color)
        datasets.append(dataset)
        return datasets

    def get_dataset(self):
        return self.datasets[0]


class PieChart(DoughnutChart):

    def __init__(self, *args, **kwargs):
        kwargs['chart_type'] = "pie"
        super().__init__(*args, **kwargs)


class BarChart(CategoryChart):

    def __init__(self, *args, **kwargs):
        if "chart_type" not in kwargs:
            kwargs["chart_type"] = "bar"
        super().__init__(*args, **kwargs)
        self.set_option("money", kwargs.get('money', False))
        self.set_option("stacked", kwargs.get('stacked', False))
        self.set_option("xAxes_name", kwargs.get('xAxes_name', "xAxes"))
        self.set_option("yAxes_name", kwargs.get('yAxes_name', 'yAxes'))

    def set_stacked(self, stacked):
        self.set_option('stacked', "true" if stacked else "false")

    def set_horizontal(self, horizontal):
        if horizontal:
            self.options["xAxes_name"] = "yAxes"
            self.options["yAxes_name"] = "xAxes"
            self.options["chart_type"] = "horizontalBar"
        else:
            self.options["xAxes_name"] = "xAxes"
            self.options["yAxes_name"] = "yAxes"
            self.options["chart_type"] = "bar"

    def get_dataset_options(self, index, color):
        default_opt = {
            "backgroundColor": "rgba(%d, %d, %d, 0.65)" % color
            # "borderColor": "rgba(%d, %d, %d, 1)" % color,
            # "borderWidth": 0.2,
        }
        if not self.options['histogram']:
            default_opt['borderColor'] = "rgba(%d, %d, %d, 1)" % color
            default_opt['borderWidth'] = 0.2
        return default_opt

    def get_dataset_configs(self):
        datasets = []
        color_generator = self.get_color_iterator()
        data = self.get_datasets()
        dataset_labels = self.get_dataset_labels()
        num = len(dataset_labels)
        for i, entry in enumerate(data):
            color = tuple(next(color_generator))
            dataset = {"data": entry}
            dataset.update(self.get_dataset_options(i, color))
            if i < num:
                dataset["label"] = dataset_labels[i]  # series labels for Chart.js
                dataset["name"] = dataset_labels[i]  # HighCharts may need this
            datasets.append(dataset)
        return datasets


class StackedBarChart(BarChart):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_stacked(True)


class HorizontalBarChart(BarChart):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_horizontal(True)


class HorizontalStackedBarChart(BarChart):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_horizontal(True)
        self.set_stacked(True)


class Histogram(BarChart):

    def __init__(self, *args, **kwargs):
        if "histogram" not in kwargs:
            kwargs["histogram"] = True
        super().__init__(*args, **kwargs)

    def build(self, data, bin_count):
        num_min = min(data)
        num_max = max(data)

        num_range = num_max - num_min
        bin_chunk = num_range / bin_count
        bin_power = math.floor(math.log10(bin_chunk))
        bin_chunk = math.ceil(bin_chunk / 10**bin_power) * 10**bin_power
        if num_min < 0:
            bin_min = math.ceil(math.fabs(num_min / bin_chunk)) * bin_chunk * -1
        else:
            bin_min = math.floor(num_min / bin_chunk) * bin_chunk
        if num_max < 0:
            bin_max = math.floor(math.fabs(num_max / bin_chunk)) * bin_chunk * -1
        else:
            bin_max = math.ceil(num_max / bin_chunk) * bin_chunk
        categories = list(range(bin_min, bin_max + bin_chunk, bin_chunk))
        self.set_option('max', categories[-2])
        self.set_option('histogram_max', categories[-1])
        bins = [0] * bin_count
        for num in data:
            for i in range(len(categories) - 1):
                if num >= categories[i] and num < categories[i + 1]:
                    if i < len(bins):
                        bins[i] += 1
        self.set_categories(categories)
        self.add_dataset(bins, 'data')


class HorizontalHistogram(Histogram):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_horizontal(True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from dataclasses import dataclass
from typing import List, Optional, Dict, Any

from .base import Block


@dataclass
class BreadcrumbItem:
    title: str
    url: str


class BreadrumbBlock(Block):
    template_name: str = 'wildewidgets/breadcrumb_block.html'
    tag: str = 'nav'
    aria_attributes: Dict[str, str] = {'label':'breadcrumb'}
    title_class: str = ""

    def __init__(self, *args, **kwargs):
        self.title_class = kwargs.pop('title_class', self.title_class)
        super().__init__(*args, **kwargs)
        self.items = []

    def add_breadcrumb(self, title, url=None):
        self.items.append(BreadcrumbItem(title, url))

    def get_context_data(self, **kwargs) -> Dict[str, Any]:
        kwargs = super().get_context_data(**kwargs)
        kwargs['items'] = self.items
        kwargs['title_class'] = self.title_class        
        return kwargs

    def flatten(self):
        return ' - '.join([item.title for item in self.items])


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from copy import copy
from typing import Union, List, Dict, Any, Optional

from django.template.loader import get_template


class Widget:
    """
    The base class form which all widgets should inherit.
    """
    title: Optional[str] = None
    icon: str = 'gear'

    def __init__(self, **kwargs):
        self.title = kwargs.pop('title', self.title)
        self.icon = kwargs.pop('icon', self.icon)
        super().__init__(**kwargs)

    def is_visible(self) -> bool:
        return True

    def get_content(self, **kwargs) -> str:
        raise NotImplementedError


class TemplateWidget(Widget):
    template_name: str

    def get_content(self, **kwargs) -> str:
        context = self.get_context_data(**kwargs)
        html_template = get_template(self.template_name)
        content = html_template.render(context)
        return content

    def __str__(self) -> str:
        return self.get_content()

    def get_context_data(self, **kwargs) -> Dict[str, Any]:
        return kwargs


class Block(TemplateWidget):
    """
    Render a single HTML element.   Example::

        Block('Hello World', tag='a', name='foo', modifier='bar', css_class='blah dah',
            attributes={'href': 'https://example.com'})

    When rendered in the template with the ``wildewdigets`` template tag, this will produce::

        <a href="https://example.com" class="foo foo--bar blah dah">Hello World</a>

    All the constructor parameters can be set in a subclass of this class as class attributes.  Parameters
    to the constructor override any defined class attributes.

    :param *blocks: any content to be rendered in the block, including other blocks.
    :type *blocks: list of strings or other blocks
    :param tag: the name of the element to use, defaults to 'div'
    :type tag: str
    :param name: This CSS class will be added to the classes to identify this element, defaults to 'block'
    :type tag: str
    :param modifier: If specified, also add a class named ``{name}--{modifier}`` to the CSS classes, defaults
        to no modifier
    :type modifier: str
    :param css_class: a string of classes to apply to the element, defaults to no classes.
    :type css_class: str
    :param css_id: Use this as the ``id`` attribute for the element, defaults to no ``id``
    :type css_id: str
    :param content: Set this as the content for the element, defaults to no content.
    :type content: str or :class:`wildewidgets.Widget`
    :param attributes: Set any additional attributes for the element as key, value pairs, defaults to no additional
        attributes.
    :type attributes: dict(str, str)
    :param data_attributes: Set ``data-`` attributes for the element, defaults to no data attributes
    :type modifier: dict(str, str)
    """
    template_name: str = "wildewidgets/block.html"

    # block is the official wildewidgets name of the block; it can't be changed by constructor
    # kwargs
    block: str = 'block'

    tag: str = 'div'
    name: Optional[str] = None
    modifier: Optional[str] = None
    css_class: Optional[str] = None
    css_id: Optional[str] = None
    content: Optional[Union[str, Widget]] = None
    attributes: Dict[str, str] = {}
    data_attributes: Dict[str, str] = {}
    aria_attributes: Dict[str, str] = {}

    def __init__(self, *blocks, tag=None, name=None, modifier=None, css_class=None, css_id=None,
                 attributes=None, data_attributes=None, aria_attributes=None):
        self._name = name if name is not None else copy(self.name)
        self._modifier = modifier if modifier is not None else self.modifier
        self._css_class = css_class if css_class is not None else self.css_class
        self._css_id = css_id if css_id is not None else self.css_id
        self._tag = tag if tag is not None else self.tag
        self._attributes = attributes if attributes is not None else copy(self.attributes)
        self._data_attributes = data_attributes if data_attributes is not None else copy(self.data_attributes)
        self._aria_attributes = aria_attributes if aria_attributes is not None else copy(self.aria_attributes)
        self.blocks = []
        for block in blocks:
            self.add_block(block)        

    def add_block(self, block):
        self.blocks.append(block)

    def get_context_data(self, **kwargs) -> Dict[str, Any]:
        context = super().get_context_data(**kwargs)
        css_class = self._css_class if self._css_class is not None else ''
        name = self._name if self._name is not None else ''
        block = self.block if self.block is not None else ''
        if name and self._modifier:
            modifier = f'{name}--{self.modifier}'
        else:
            modifier = ''
        context['tag'] = self._tag
        context['block_name'] = block
        context['name'] = name
        context['css_classes'] = ' '.join([name, modifier, block, css_class]).strip()
        context['css_id'] = self._css_id
        context['blocks'] = self.blocks
        context['attributes'] = self._attributes
        context['data_attributes'] = self._data_attributes
        context['aria_attributes'] = self._aria_attributes
        return context


class WidgetStream(Block):
    template_name: str = 'wildewidgets/widget_stream.html'
    block: str = 'widget-stream'
    widgets: List[Widget] = []

    def __init__(self, **kwargs):
        self._widgets: List[Widget] = copy(kwargs.pop('widgets', self.widgets))
        super().__init__(**kwargs)

    @property
    def is_empty(self) -> bool:
        return len(self._widgets) == 0

    def add_widget(self, widget, **kwargs):
        wrapper = Block(widget, **kwargs)
        wrapper.block = f"{self.block}__widget"
        self._widgets.append(wrapper)

    def get_context_data(self, **kwargs) -> Dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context['widgets'] = self._widgets
        return context
